/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package clique;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

public class Main {
	private Map<String, Map<String, Double>> map = new HashMap<String, Map<String, Double>>();
	private List<String> hosts = new ArrayList<String>();
	private Vector<String> firstClique = new Vector<String>();
	private int sizeClique = 60;

	public static void main(String[] args) throws IOException {
		new Main();
	}

	public Main() throws IOException {
		parseDatasetFile();
		doCliqueBT(new Vector<String>(), 0);
	}

	/**
	 * Backtracking clique algorithm.
	 * 
	 * Source: http://www.ibluemojo.com/school/clique_algorithm.html
	 * 
	 * Date: August 16, 2012
	 * 
	 * @param A
	 * @param j
	 * @throws IOException
	 */
	private void doCliqueBT(Vector<String> A, int j) throws IOException {
		// If j is equal to size of clique, k, then A is k-clique in the graph
		if (j == sizeClique) {
			if (firstClique.isEmpty())
				firstClique = A;
			writeClique(A);
			sizeClique++;
			// return;
		} else {
			j++;

			// Sj is the set of all candidate vectors for j-clique
			ArrayList<Vector<String>> Sj = new ArrayList<Vector<String>>();

			if (j <= sizeClique)
				Sj = getCandidates(A);

			if (!Sj.isEmpty()) {
				// For each candidate vector in Sj,
				// recursively do backtracking for k-clique
				for (int i = 0; i < Sj.size(); i++) {
					// Vector<String> a = (Vector<String>) Sj.get(i);
					doCliqueBT(Sj.get(i), j);
				}
			}
		}
	}

	/**
	 * Create list of vectors, representing possible "clique"-sets
	 * 
	 * @param A
	 * @return
	 */
	private ArrayList<Vector<String>> getCandidates(Vector<String> A) {
		// The set of all candidate vectors for q-clique
		ArrayList<Vector<String>> candidates = new ArrayList<Vector<String>>();

		if (A.isEmpty()) {
			for (String host : hosts) {
				Vector<String> sj = new Vector<String>(1);
				sj.add(host);
				candidates.add(sj);
			}
		} else {
			int idx = hosts.indexOf((String) A.lastElement());
			for (int i = idx + 1; i < hosts.size(); i++) {
				String j = hosts.get(i);

				// Check if vertex 'j' is adjacent to all vertices in A
				boolean allConnected = true;
				for (String v : A) {
					if (!isConnected(v, j)) {
						allConnected = false;
						break;
					}
				}

				if (allConnected) {
					Vector<String> sj = new Vector<String>(A);
					sj.add(j);
					candidates.add(sj);
				}
			}
		}
		return candidates;
	}

	/**
	 * Write "clique"-set of hosts to a file
	 * 
	 * @param A
	 * @throws IOException
	 */
	private void writeClique(Vector<String> A) throws IOException {
		FileWriter writer = new FileWriter("../BandwidthPredictionSim/data/clique.txt");
		for (int i = 0; i < A.size(); i++) {
			String v = (String) A.get(i);
			writer.write(v + "\n");
		}
		writer.close();
		System.out.println("Maximal Clique found, containing " + sizeClique + " nodes.");
	}

	/**
	 * Check if two hosts are connected
	 * 
	 * @param a
	 * @param b
	 * @return
	 */
	private boolean isConnected(String a, String b) {
		if (map.containsKey(a)) {
			Map<String, Double> amap = map.get(a);
			if (amap.containsKey(b)) {
				if (map.containsKey(b)) {
					Map<String, Double> bmap = map.get(b);
					if (bmap.containsKey(a)) {
						return true;
					}
				}
			}
		}
		return false;
	}

	/**
	 * Parse dataset file
	 * 
	 * @throws NumberFormatException
	 * @throws IOException
	 */
	private void parseDatasetFile() throws NumberFormatException, IOException {
		// read file
		BufferedReader reader = new BufferedReader(new FileReader("data/hp-planetlab-bandwidth-snapshot.csv"));
		String line;
		while ((line = reader.readLine()) != null) {
			line = line.trim();

			// skip comments
			if (line.startsWith("#"))
				continue;

			// parse
			String[] tokens = line.split(",");
			if (tokens.length < 5)
				continue;

			String a = tokens[0];
			String b = tokens[1];

			if (tokens[4].equals("N/A"))
				continue;

			double bw = Double.parseDouble(tokens[4]);

			// store
			addHost(a);
			addHost(b);
			add(a, b, bw);
		}

		reader.close();

		System.err.println("Parsing complete.");
	}

	/**
	 * Add host to list of hosts
	 * 
	 * @param host
	 */
	private void addHost(String host) {
		if (!hosts.contains(host))
			hosts.add(host);
	}

	/**
	 * Add a link bandwidth between two hosts to the map
	 * 
	 * @param a
	 * @param b
	 * @param bw
	 */
	private void add(String a, String b, double bw) {
		if (!map.containsKey(a))
			map.put(a, new HashMap<String, Double>());
		map.get(a).put(b, bw);
	}

}
