/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import sequoia.Master.Output;
import sequoia.Master.Topologies;
import sequoia.Master.Trees;
import sequoia.topology.*;
import sequoia.tree.*;

/**
 * helper functions
 */
public class Helper {

	// max length of hostnames for output
	private static final int HOSTSIZE = 10;

	/**
	 * returns a prediction tree, containing the nodes of the given topology
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static IPredictionTree getTree(Trees t, ITopology topology) {
		IPredictionTree tree = null;

		switch (t) {
		case SEQUOIA:
			tree = new TreeKSequoia(topology);
			break;
		case SONG:
			tree = new TreeSong(topology);
			break;
		case MYTREE:
			tree = new TreeMy(topology);
			break;
		default:
			throw new RuntimeException("Invalid tree selection");
		}

		// embed all hosts of given topology
		List<String> topHosts = topology.getAllHosts();
		tree.embedAll(topHosts);
		
		// give some feedback
		if(Master.cli_verbose) {
			if(t.equals(Trees.MYTREE))
				System.err.println("Embedded "+((tree.getHosts().size()-1)/2+1)+"/"+topology.getAllHosts().size()+" hosts.");
			else
				System.err.println("Embedded "+tree.getHosts().size()+"/"+topology.getAllHosts().size()+" hosts.");
		}
		
		return tree;
	}

	/**
	 * returns a topology instance
	 */
	public static ITopology getTopology(Topologies t) {
		switch (t) {
		case PERFECT:
			return new TopPerfectTreeness();
		case REALKNOWN:
			return new TopRealKnown();
		case DSLCONNECTIONS:
			return new TopDSL();
		case PLANETLAB:
			TopPlanetLab plt = new TopPlanetLab();
			if (Master.cli_useclique) {
				plt.readCliqueFile(Master.CLIQUE_FILE);
			}
			if(Master.cli_verbose) {
				System.err.println("topology size: " + plt.getAllHosts().size());
			}
			return plt;
		case SUBSETPL: 
			return new TopPlanetLab("16-clique.dat");
		}
		return null;
	}

	/**
	 * table visualization 
	 */
	@SuppressWarnings("rawtypes")
	public static String matrix(IPredictionTree tree, Output cli_output, boolean csv) {
		ITopology top = tree.getTopology();
		String HFormat = "%" + String.valueOf(HOSTSIZE) + "s";
		String FORMAT = csv ? HFormat + ";" : HFormat + "|";
		String out = "";
		out += "Topology: " + top.getTopologyName() + " " + cli_output + "\n";
		out += String.format(FORMAT, "from/to");
		List<String> hosts = new ArrayList<String>(top.getAllHosts());
		Collections.sort(hosts);
		for (String host : hosts) {
			out += String.format(FORMAT, formatHostname(host));
		}
		out += "\n";

		for (String origin : hosts) {
			out += String.format(FORMAT, formatHostname(origin));
			for (String sink : hosts) {
				if (origin.equals(sink)) {
					out += String.format(FORMAT, "*");
				} else {
					double value = 0;
					if (cli_output.equals(Output.MEASURED)) {
						value = 0;//top.dist(origin, sink);
					} else if (cli_output.equals(Output.PREDICTED)) {
						value = tree.predictBandwidth(origin, sink);
					} else if (cli_output.equals(Output.DIFF)) {
						int m = 0;//top.dist(origin, sink);
						double p = tree.predictBandwidth(origin, sink);
						value = Math.abs(m - p);
					} else if (cli_output.equals(Output.RELERROR)) {
						int m = 0;//top.dist(origin, sink);
						double p = tree.predictBandwidth(origin, sink);
						value = Math.abs(m - p) / m;
					}

					out += String.format(FORMAT, format(value));
				}
			}
			out += "\n";
		}

		return out;
	}

	/**
	 * hostname formatter for output (shortens long hostnames)
	 */
	private static String formatHostname(String host) {
		if (host.length() > 10) {
			return host.substring(0, HOSTSIZE);
		} else {
			return host;
		}
	}

	/**
	 * double value formatter for output
	 */
	private static String format(double value) {
		DecimalFormatSymbols unusualSymbols = new DecimalFormatSymbols();
		unusualSymbols.setDecimalSeparator(',');

		String strange = "###.#";
		DecimalFormat weirdFormatter = new DecimalFormat(strange, unusualSymbols);

		String bizarre = weirdFormatter.format(value);
		return bizarre;
	}

	/**
	 * round v to 2 decimal places
	 */
	public static double round(double v) {
		return round(v, 2);
	}

	/**
	 * round v to s decimal places
	 */
	public static double round(double v, int s) {
		s = s < 1 ? 1 : s;
		double factor = Math.pow(10, s);
		return Math.round(v * factor) / factor;
	}

	/**
	 * list of doubles to double[]
	 */
	public static double[] toArray(List<Double> list) {
		int size = list.size();
		double[] array = new double[size];
		for (int i = 0; i < size; i++) {
			array[i] = list.get(i);
		}
		return array;
	}

	/**
	 * write CDF to file
	 */
	public static void writeToFile(String path, String filename, String string) throws IOException {
		path = path.endsWith("/") ? path : path + "/";
		File folder = new File(path);
		folder.mkdir();

		File file = new File(filename);
		FileWriter wri = new FileWriter(path + file);

		wri.write(string);
		wri.close();
	}

}
