/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.topology;

import sequoia.Master;

/**
 * DSLConnections topology. Simulates various servers (symmetric bandwidth) and
 * various private hosts (asymmetric bandwidth)
 * 
 * @author stefan
 * 
 */
public class TopDSL extends HostConnectionTopology {

	// relation of downstream and upstream
	private static final double UD_RATIO = 0.15;

	// maximal downstream capacity (kbps)
	private static final double CLIENT_DOWNSTREAM_BW = 100000;
	private static final double SERVER_DOWNSTREAM_BW = 100000;

	// amount of hosts/servers created
	private static int HOST_COUNT = 40 * Master.cli_nodescale;
	private static int SERVER_COUNT = 20 * Master.cli_nodescale;

	// amount of noise added to server bandwidth
	private static final double NOISE = 0.1;

	public TopDSL() {

		int totalHostCount = 0;

		// handle symmetry parameter
		if (Master.cli_symratio > -1) {
			// validate given value
			if (Master.cli_symratio < 0 && Master.cli_symratio > 1) {
				System.err.println("Invalid symmetry value (0.0 - 0.1 only)");
			} else {
				// default 60 Hosts, scalable by CLI parameter.
				totalHostCount = 60 * Master.cli_nodescale;
				HOST_COUNT = (int) Math.round(totalHostCount * Master.cli_symratio);
				SERVER_COUNT = Math.round(totalHostCount - HOST_COUNT);
			}
		}

		setDescription("Simulates DSL Connections, hosts that have asymmetric bandwidth ratio of 10:1 for downstream/upstream.");
		for (int i = 0; i < HOST_COUNT; i++) {
			int down = (int) Math.round(Master.rand.nextDouble() * CLIENT_DOWNSTREAM_BW);
			int up = (int) Math.round(UD_RATIO * down);
			addHostBandwLimits(i + "PrivHost", down, up);
		}

		for (int i = 0; i < SERVER_COUNT; i++) {
			// random bandwidth noise +- 10%
			double amount = SERVER_DOWNSTREAM_BW * NOISE;
			int bw = (int) Math.round(amount / 2 - Master.rand.nextDouble() * amount);
			bw = (int) (SERVER_DOWNSTREAM_BW + bw);
			addHostBandwLimits(i + "Server", bw, bw);
		}

		genTable();
		shuffle();
	}

}
