/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.topology;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import sequoia.*;

/**
 * Abstract topology, implementing common methods for all topologies
 */
abstract public class Topology implements ITopology {

	protected Random rnd = Master.rand;
	protected String _description;
	List<String> _hosts = new ArrayList<String>();
	Map<String, Map<String, Integer>> _bandwidth = new HashMap<String, Map<String, Integer>>();
	Map<String, List<String>> _measured = new HashMap<String, List<String>>();

	protected void setDescription(String s) {
		_description = s;
	}

	protected void addLinkBandwidth(String a, String b, int val) {
		if(val == 0)
			throw new RuntimeException("Adding zero bandwidth not allowed.");
		if (!_bandwidth.containsKey(a)) {
			_bandwidth.put(a, new HashMap<String, Integer>());
		}
		_bandwidth.get(a).put(b, val);
	}

	public List<String> getAllHosts() {
		List<String> list = new ArrayList<String>(_hosts);
		Collections.copy(list, _hosts);
		return list;
	}

	@Override
	public String getTopologyName() {
		return this.getClass().getSimpleName();
	}

	public void shuffle() {
		// shuffle hosts, for random embedding sequence
		Collections.shuffle(_hosts, rnd);
	}

	public int getBandwidth(String a, String b, boolean countMeasurement) {
		if (_bandwidth.containsKey(a))
		{
			Map<String, Integer> amap = _bandwidth.get(a);
			if (amap.containsKey(b)) {
				int v = amap.get(b);
				if(v < 0) 
					throw new RuntimeException("Negative bandwidth value in topology.");
				if(v > Master.CONST) 
					throw new RuntimeException("Bandwidth value in topology greater than CONST");
				if(v == 0) 
					System.err.println("Topology returns bandwidth="+v);
				if(countMeasurement) 
					countMeasurement(a,b);
				return v;
			}
		}
		return 0;
	}
	
	public int getBandwidth(String a, String b) {
		return getBandwidth(a, b, true);		
	}

	private void countMeasurement(String a, String b) {
		if(!_measured.containsKey(a)) 
			_measured.put(a, new ArrayList<String>());
		if(!_measured.get(a).contains(b))
			_measured.get(a).add(b);
	}

	protected boolean bandwidthValExists(String a, String b) {
		return _bandwidth.containsKey(a) && _bandwidth.get(a).containsKey(b);
	}

	/**
	 * Remember measurement, to generate a "measured" list later
	 */
	protected void rememberHostname(String hostname) {
		if (!_hosts.contains(hostname))
			_hosts.add(hostname);
	}

	public String toString() {
		// copy and sort hosts.
		ArrayList<String> hosts = new ArrayList<String>(_hosts);
		Collections.copy(hosts, _hosts);
		Collections.sort(hosts);
	
		// generate output
		String s = "Scenery: " + _description + "\n";
		s += "\n";
		s += "  von/nach|";
		String underline = "";
		for (String a : hosts) {
			s += String.format("%10s|", a);
			underline += "------------";
		}
		s += "\n";
		s += underline + "\n";
	
		for (String a : hosts) {
			s += String.format("%10s|", a);
			for (String b : hosts) {
				if (a.equals(b)) {
					s += String.format("%10s|", "*");
				} else {
					s += String.format("%10s|", getBandwidth(a, b, false));
				}
			}
			s += "\n";
		}
		return s;
	}

	public CDF getSymmetry() {
		List<Double> list = new ArrayList<Double>();
		List<String> hosts = getAllHosts();
		for (int a = 0; a < hosts.size(); a++) {
			for (int b = a + 1; b < hosts.size(); b++) {
				String aa = hosts.get(a);
				String bb = hosts.get(b);
				double forth = getBandwidth(aa, bb);
				double back = getBandwidth(bb, aa);
				if(forth == 0 || back == 0)
					continue;
				
				// pathGuru way of calculating symmetry 
				double symm = Math.abs( back - forth ) / ( back + forth ) ;
				list.add(symm);
			}
		}
		CDF cdf = new CDF(list,Master.CDF_ERROR_MAX);
		return cdf;
	}

	@Override
	public CDF getTreeness() {
		List<String> hosts = getAllHosts();
		List<Double> fpces = new ArrayList<Double>();
	
		// for every set of four nodes, calculate epsilon (4PC-e)
		for (int a = 0; a < hosts.size(); a++) {
			for (int b = a + 1; b < hosts.size(); b++) {
				for (int c = b + 1; c < hosts.size(); c++) {
					for (int d = c + 1; d < hosts.size(); d++) {
						String[] set = { hosts.get(a), hosts.get(b), hosts.get(c), hosts.get(d) };
	
						double fpce = fPCe(set);
						if (Double.isInfinite(fpce) || Double.isNaN(fpce))
							continue;
	
						if (Master.TREENESSLIMIT > 0 && fpce > Master.TREENESSLIMIT)
							fpce = Master.TREENESSLIMIT;
	
						fpces.add(fpce);
					}
				}
			}
		}
	
		CDF cdf = new CDF(fpces,Master.CDF_ERROR_MAX);
		return cdf;
	}

	/**
	 * get distance, averaging both directions for 4PC-epsilon calculation
	 */
	private double fpce_get(String[] hosts, int a, int b) {
		double there = getBandwidth(hosts[a], hosts[b]);
		double back = getBandwidth(hosts[b], hosts[a]);
		if (there == 0 && back == 0)
			throw new RuntimeException("MissingMeasurement");
		if(there == 0 && back != 0)
			there = back;
		else if(back == 0 && there != 0)
			back = there;
		double value = (there + back) / 2;
		value = Master.CONST - value;
		return value;
	}

	/**
	 * Calculate 4PC-epsilon for four hosts
	 */
	private double fPCe(String[] hosts) {
		// coefficients
		double dwz;
		double dxy;
		double dwy;
		double dxz;
		double dwx;
		double dyz;
		try {
			dwz = fpce_get(hosts, 0, 3);
			dxy = fpce_get(hosts, 1, 2);
			dwy = fpce_get(hosts, 0, 2);
			dxz = fpce_get(hosts, 1, 3);
			dwx = fpce_get(hosts, 0, 1);
			dyz = fpce_get(hosts, 2, 3);
		} catch (Exception e) {
			return -1;
		}
	
		// sums
		double sum1 = dwx + dyz;
		double sum2 = dwy + dxz;
		double sum3 = dwz + dxy;
	
		// cases
		double first = 0;
		double second = 0;
		double min_third = 0;
	
		if (sum1 >= sum2 && sum1 >= sum3) {
			first = sum1;
			if (sum2 >= sum3) {
				second = sum2;
				min_third = min(dwz, dxy);
			} else {
				second = sum3;
				min_third = min(dwy, dxz);
			}
		} else if (sum2 >= sum1 && sum2 >= sum3) {
			first = sum2;
			if (sum1 >= sum3) {
				second = sum1;
				min_third = min(dwz, dxy);
			} else {
				second = sum3;
				min_third = min(dwx, dyz);
			}
		} else {
			first = sum3;
			if (sum2 >= sum1) {
				second = sum2;
				min_third = min(dwx, dyz);
			} else {
				second = sum1;
				min_third = min(dwy, dxz);
			}
		}
	
		// depends on: the two greater sums, min coefficient of the smallest sum
		double e = (first - second) / (2 * min_third);
		return e;
	}

	private double min(double a, double b) {
		return a > b ? b : a;
	}
	
	public int getMeasCount() {
		int sum = 0;
		for(List<String> list : _measured.values())
			sum+=list.size();
		return sum;
	}

}
