/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.tree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import sequoia.Master;
import sequoia.Master.Trees;

/**
 * Distance label implementation, containing the distance calculation algorithm
 * by Song et al.
 * 
 * http://www.cs.umd.edu/Grad/scholarlypapers/papers/SukhyunSong.pdf, August 15, 2012
 * 
 * @author stefan
 * 
 * @param <T>
 */
public class DistanceLabel<T> {
	/**
	 * lists for hosts, offsets and distances. triple(0) = hosts(0), offsets(0),
	 * distances(0).
	 */
	public List<T> hosts = new ArrayList<T>();
	public List<Double> offsets = new ArrayList<Double>();
	public List<Double> distances = new ArrayList<Double>();

	/** add a new triple to the label */
	public void add(T h, double i, double j) {
		hosts.add(h);
		offsets.add(i);
		distances.add(j);
	}

	/** update values of the last triple of the label */
	public void updateLast(double offset, double distance) {
		int size = hosts.size();
		offsets.remove(size - 1);
		offsets.add(offset);
		distances.remove(size - 1);
		distances.add(distance);
	}

	/** calculate distance between two labels */
	public double distanceTo(DistanceLabel<T> label) {
		DistanceLabel<T> p = this;
		DistanceLabel<T> q = label;

		/** lowest-level common anchor "z" in label p and label q */
		// example labels: (just an anchor list)
		// p = a,b,c,d,e,f
		// q = g,h,c,d,i,j,k,l
		// lowest-level common anchor = d, not c

		int zidx_p = 0;
		int zidx_q = 0;

		// iteration through all anchors in reverse direction
		T z_in_p;
		T z_in_q;
		outerloop: for (zidx_p = p.hosts.size() - 1; zidx_p >= 0; zidx_p--) {
			for (zidx_q = q.hosts.size() - 1; zidx_q >= 0; zidx_q--) {
				z_in_p = (T) p.hosts.get(zidx_p);
				z_in_q = (T) q.hosts.get(zidx_q);
				if (z_in_p.equals(z_in_q)) {
					break outerloop;
				}
			}
		}

		// check valid list size
		if (zidx_p >= p.hosts.size() && zidx_p >= p.offsets.size() && zidx_p >= p.distances.size()) {
			throw new RuntimeException("index out of bounds (p)");
		}

		if (zidx_q >= q.hosts.size() && zidx_q >= q.offsets.size() && zidx_q >= q.distances.size()) {
			throw new RuntimeException("index out of bounds (q)");
		}

		/** node "x" whose anchor node is z in p */
		/** node "y" whose anchor node is z in q */
		/** x´s virtual anchor node "tx" */
		/** y´s virtual anchor node "ty" */

		/** distance = | d_T(z,tx) - d_T(z,ty) | + d_T(tx,x) + d_T(ty,y) */
		// d_T(z,tx) - d_T(z,ty) => offset difference
		double d_z_tx = (Double) p.offsets.get(zidx_p);
		double d_z_ty = (Double) q.offsets.get(zidx_q);

		double offset_diff = Math.abs(d_z_tx - d_z_ty);
		if (d_z_tx < 0 && d_z_ty < 0) {
			/**
			 * My simple bugfix patch for the original algorithm, in order to
			 * achieve correct distance calculation for negative offset values
			 * Thus, only used for MyTREE.
			 */
			if (Master.cli_tree.equals(Trees.MYTREE)) {
				if (Master.cli_dlabelmod) {
					// benefit?
					offset_diff *= -1;
				}
			}
		}

		// d_T(tx,x) => x distance
		double x_distance = (Double) p.distances.get(zidx_p);

		// d_T(ty,y) => y distance
		double y_distance = (Double) q.distances.get(zidx_q);

		double distance = offset_diff + x_distance + y_distance; // */

		/** foreach node "a" in sublabel (x -> ... -> p) and (y -> ... -> q) */
		for (int a = zidx_p + 1; a < p.offsets.size(); a++) {
			/** if "a" != x && "a" != y */
			// "if" not needed, because of start index "+2"
			/** "a2" = a´s anchor node */
			/** "ta" = a´s virtual anchor node */
			/** distance = distance - d_T(a2,ta) + d_T(ta,a) */
			double d_a2_ta = (Double) p.offsets.get(a); // offset
			double d_ta_a = (Double) p.distances.get(a); // distance

			distance = distance - d_a2_ta + d_ta_a;
		}

		for (int a = zidx_q + 1; a < q.offsets.size(); a++) {
			/** if "a" != x && "a" != y */
			// "if" not needed, because of start index "+2"
			/** "a2" = a´s anchor node */
			/** "ta" = a´s virtual anchor node */
			/** distance = distance - d_T(a2,ta) + d_T(ta,a) */
			double d_a2_ta = (Double) q.offsets.get(a); // offset
			double d_ta_a = (Double) q.distances.get(a); // distance

			distance = distance - d_a2_ta + d_ta_a;
		}

		// FIXME: that should not be done for Sequoia i think!
		return Math.abs(distance);
	}

	/** returns immediate anchor of this */
	public int getAnchor() {
		if (hosts.size() >= 2) {
			return hosts.size() - 2;
		} else {
			return hosts.size() - 1;
		}
	}

	/** returns a copy of this label */
	public DistanceLabel<T> clone() {
		DistanceLabel<T> ndl = new DistanceLabel<T>();
		ndl.hosts = new ArrayList<T>(hosts);
		ndl.offsets = new ArrayList<Double>(offsets);
		ndl.distances = new ArrayList<Double>(distances);
		Collections.copy(ndl.hosts, hosts);
		Collections.copy(ndl.offsets, offsets);
		Collections.copy(ndl.distances, distances);
		return ndl;
	}

	/** string representation of a label */
	public String toString() {
		String ret = "";
		for (int i = 0; i < hosts.size(); i++) {
			ret += String.format("(%5s\t%5s\t%5s),", hosts.get(i).toString().substring(0, 8), offsets.get(i), distances.get(i));
		}
		return ret;
	}

}
