/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.tree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;

import sequoia.Master;
import sequoia.topology.ITopology;

public class TreeKSequoia implements IPredictionTree<Host> {

	private static final int _k = Master.cli_sequoiak;
	private List<TreeSequoia> _trees = new ArrayList<TreeSequoia>();
	private List<Host> _hosts = new ArrayList<Host>();
	private ITopology _top;
	
	public TreeKSequoia(ITopology top) {
		_top = top;
		for(int i=0;i<_k;i++)
			_trees.add(new TreeSequoia(top));
	}
	
	@Override
	public void embed(String host) {
		throw new RuntimeException("Not implemented.");
	}
	
	@Override
	public void embedAll(List<String> hosts) {
		// embed all in different order
		for(int i=0;i<_k;i++) {
			for(String host : hosts) {
				try {
				_trees.get(i).embed(host);
				} catch(Exception e) {
					System.err.println("Node skipped: "+e.toString());
				}
			}
			Collections.shuffle(hosts, Master.rand);
		}
	}
	
	@Override
	public List<String> getHostnames() {
		List<String> hostnames = new ArrayList<String>();
		for(Host host : getHosts())
			hostnames.add(host.hostname);
		return hostnames;
	}
	
	@Override
	public List<Host> getHosts() {
		for(TreeSequoia tree : _trees) {
			for(Host host : tree._hosts) {
				if(!_hosts.contains(host))
					_hosts.add(host);
			}
		}
		return _hosts;
	}
	
	private Host getHost(String hostname) {
		List<Host> hosts = getHosts();
		for(Host host : hosts) {
			if(host.hostname.equals(hostname))
				return host;
		}
		throw new RuntimeException("Host not found "+hostname);
	}
	
	@Override
	public DistanceLabel<Host> getDLabel(String hostname) {
		Host host = getHost(hostname);
		return host.label;
	}
	@Override
	public double predictBandwidth(String a, String b) {
		DescriptiveStatistics descrStats = new DescriptiveStatistics();
		for(TreeSequoia tree : _trees) {
			double predictedBandwidth = tree.predictBandwidth(a, b);
			if(!Double.isInfinite(predictedBandwidth))
				descrStats.addValue(predictedBandwidth);
		}
		double medianBandwidth = descrStats.getPercentile(50);
		return medianBandwidth;
	}
	
	@Override
	public double transform(double v) {
		throw new RuntimeException("Not implemented");
	}
	
	@Override
	public ITopology getTopology() {
		return _top;
	}

	@Override
	public double getMeasurementRatio(){
		List<String> hostnames = getHostnames();
		Map<String, List<String>> measured = new HashMap<String, List<String>>();
		for(String source : hostnames) {
			for(String destination: hostnames) {
				for(TreeSequoia sequoia : _trees) {
					if(sequoia._measured.containsKey(source)) {
						List<String> sourceList = sequoia._measured.get(source);
						if(sourceList.contains(destination)) {
							// count measurement
							if(!measured.containsKey(source))
								measured.put(source, new ArrayList<String>());
							if(!measured.get(source).contains(destination))
								measured.get(source).add(destination);
							break;
						}
					}
				}
			}
		}
		double measurements = 0;
		for(String source : measured.keySet())
			measurements += measured.get(source).size(); 
		double size = hostnames.size();
		double links = (size * (size - 1));
		double ratio = measurements / links;
		return ratio;
	}

}
